class Error(Exception):
    pass


class BackendError(Error):
    def __init__(self, message, server):
        self.message = message
        self.server = server

    def __str__(self):
        return "({}:{}) - {}" % (
            self.server.hostname, self.server.port, self.message
        )


class BackendConnectionError(BackendError):
    pass


class BackendIntegrityError(BackendError):
    pass


class TimeTravelNotAllowed(BackendError):
    pass


class BackendMaxDriftError(BackendError):
    pass


class ResourceError(Error):
    pass


class ResourceNotFound(ResourceError):
    pass



from collections import namedtuple

MappedPermission = namedtuple('MappedPermission',
                              ['permission', 'argument', 'granted_on', 'distance', 'path'])


class ResourceDict(dict):
    def __call__(self, direct=False, roles=None):
        if isinstance(roles, basestring):
            roles = [roles]
        new_dict = {}
        for key, value in self.iteritems():
            if direct and value.get("distance", 0) != 1:
                continue
            if roles and value.get("rolename") not in roles:
                continue
            new_dict[key] = value
        return new_dict


class Group(object):
    def __init__(self, groups, users, subgroups, permissions):
        self.groups = ResourceDict(groups)
        self.users = ResourceDict(users)
        self.subgroups = ResourceDict(subgroups)
        self.permissions = [
            MappedPermission(**permission) for permission in permissions
        ]

    @classmethod
    def from_payload(cls, payload):
        return cls(
            payload["data"]["groups"],
            payload["data"]["users"],
            payload["data"]["subgroups"],
            payload["data"]["permissions"],
        )


class User(object):
    def __init__(self, groups, public_keys, permissions):
        self.groups = ResourceDict(groups)
        self.public_keys = public_keys
        self.permissions = [
            MappedPermission(**permission) for permission in permissions
        ]

    @classmethod
    def from_payload(cls, payload):
        return cls(
            payload["data"]["groups"],
            payload["data"]["user"]["public_keys"],
            payload["data"]["permissions"]
        )


class Permission(object):
    def __init__(self, groups):
        self.groups = {
            groupname: Group.from_payload({"data": groups[groupname]})
            for groupname in groups
        }

    @classmethod
    def from_payload(cls, payload):
        return cls(
            payload["data"]["groups"],
        )



from clowncar.backends import Backends
import json
from threading import Lock
import time
from tornado.httpclient import HTTPClient, HTTPError

from . import exc
from .collations import Users, Groups, Permissions


class Groupy(object):
    def __init__(self, servers, partition_key=None, timeout=3,
                 allow_time_travel=False, checkpoint=0, checkpoint_time=0,
                 max_drift=600, mark_bad_timeout=60, max_backend_tries=5):

        self._lock = Lock()
        self.timeout = timeout
        self.backends = Backends(servers, partition_key)

        self.checkpoint = checkpoint
        self.checkpoint = checkpoint_time

        self.allow_time_travel = allow_time_travel
        self.max_drift = max_drift
        self.mark_bad_timeout = mark_bad_timeout
        self.max_backend_tries = max_backend_tries

        self.users = Users(self, "users")
        self.groups = Groups(self, "groups")
        self.permissions = Permissions(self, "permissions")

    def _try_get(self, path):
        for idx in range(self.max_backend_tries):
            try:
                return self._get(path)
            except exc.BackendConnectionError as err:
                self.backends.mark_dead(err.server, self.mark_bad_timeout)
        raise exc.BackendConnectionError(
            "Tried {} servers, all failed.".format(self.max_backend_tries),
            err.server
        )

    def _get(self, path):
        http_client = HTTPClient()
        server = self.backends.server
        url = "http://{}:{}{}".format(server.hostname, server.port, path)
        try:
            out = json.loads(http_client.fetch(url, **{
                "connect_timeout": self.timeout,
                "request_timeout": self.timeout,
            }).body)
        except HTTPError as err:
            if err.code == 599:
                raise exc.BackendConnectionError(err.message, server)
            try:
                out = json.loads(err.response.body)
                if "status" not in out:
                    raise exc.BackendIntegrityError(err.message, server)
            except (ValueError, TypeError):
                raise exc.BackendIntegrityError(err.message, server)

        now = time.time()
        drift = now - out["checkpoint_time"]
        if self.max_drift is not None and self.max_drift > abs(drift):
            raise exc.BackendMaxDriftError(
                "Backend last checkpoint stale by {} seconds.".format(drift),
                server
            )

        with self._lock:
            new_checkpoint = out["checkpoint"]
            old_checkpoint = self.checkpoint
            if new_checkpoint < old_checkpoint and not self.allow_time_travel:
                raise exc.TimeTravelNotAllowed(
                    "Received checkpoint of {} when previously {}".format(
                        new_checkpoint, old_checkpoint
                    ), server
                )
            self.checkpoint = new_checkpoint

        return out



from . import exc
from .resources import User, Group, Permission


class Collection(object):
    def __init__(self, client, name):
        self.client = client
        self.name = name

    def _get(self, resource=None):
        path = "/{}".format(self.name)
        if resource:
            path += "/{}".format(resource)
        response = self.client._try_get(path)
        self._check_error(response)
        return response

    def _check_error(self, response):
        if response["status"] == "ok":
            return

        combined = []
        for error in response["errors"]:
            code, message = error["code"], error["message"]
            if code == 404:
                raise exc.ResourceNotFound(message)
            combined.append("{}: {}".format(code, message))

        raise exc.ResourceError(", ".join(combined))

    def __iter__(self):
        response = self._get()
        for resource in response["data"][self.name]:
            yield resource

    def get(self, resource):
        return self.resource.from_payload(self._get(resource))


class Groups(Collection):
    resource = Group


class Users(Collection):
    resource = User


class Permissions(Collection):
    resource = Permission
